--[[============================================================
--=
--=  App Environment
--=
--=-------------------------------------------------------------
--=
--=  MyHappyList - manage your AniDB MyList
--=  - Written by Marcus 'ReFreezed' Thunström
--=  - MIT License (See main.lua)
--=
--==============================================================

	pause, unpause, isPaused
	setSetting, loadSettings, saveSettings, scheduleSaveSettings, scheduleSaveSettingsIfNeeded, setSettingsChanged, restoreDefaultSettings
	setStatusText

	addFileInfo, removeFileInfo, removeSelectedFileInfos
	anyFileInfos, noFileInfos
	checkFileInfos
	eachFileInfoByRow, getFileInfosByRows, getSelectedFileInfos
	getFileInfoByRow, getFileRow
	getFileStatus, getFileViewed
	openContainingFolder
	saveFileInfos, loadFileInfos
	setFileInfo, softUpdateFileInfo
	updateFileList

--============================================================]]

FILE_INFO_DONT_SERIALIZE       = newSet{"name","folder","isHashing"}

FILE_COLUMN_FILE               = 1
FILE_COLUMN_FOLDER             = 2
FILE_COLUMN_SIZE               = 3
FILE_COLUMN_VIEWED             = 4
FILE_COLUMN_STATUS             = 5
FILE_COLUMN_LAST               = 5

MYLIST_STATUS_UNKNOWN          = 0
MYLIST_STATUS_NO               = 1
MYLIST_STATUS_YES              = 2
MYLIST_STATUS_INVALID          = 3 -- AniDB don't know of the file (ed2k+size missing).  @Incomplete: Use this!

STATUS_BAR_FIELD_MESSAGE_QUEUE = 1

DEFAULT_APP_SETTINGS = {
	autoAddToMylist        = true,
	autoHash               = true,
	autoRemoveDeletedFiles = false,
	truncateFolders        = true,

	movieExtensions = {"avi","flv","mkv","mov","mp4","mpeg","mpg","ogm","ogv","swf","webm","wmv"},

	mylistDefaults = {
		state   = MYLIST_STATE_INTERNAL_STORAGE,
		viewed  = nil, -- bool
		source  = nil, -- string
		storage = nil, -- string
		other   = nil, -- string (newlines allowed)
	},

	windowSizeX     = -1,
	windowSizeY     = -1,
	windowPositionX = -1,
	windowPositionY = -1,
	windowMaximized = false,

	["fileColumnWidth"..FILE_COLUMN_FILE]   = 500,
	["fileColumnWidth"..FILE_COLUMN_FOLDER] = 500,
	["fileColumnWidth"..FILE_COLUMN_SIZE]   = 80,
	["fileColumnWidth"..FILE_COLUMN_VIEWED] = 80,
	["fileColumnWidth"..FILE_COLUMN_STATUS] = 120,

	language              = "en-US",
	serverResponseTimeout = DEFAULT_SERVER_RESPONSE_TIMEOUT,

	fileListScroll    = 0,
	fileListSelection = {0},
}



anidb                          = nil
appIcons                       = nil
eventQueue                     = nil
fontTitle                      = nil

topFrame                       = nil
statusBar                      = nil
statusBarProgressGauge         = nil
topPanel                       = nil
loginButton                    = nil
fileList                       = nil

fileInfos                      = {}
lastFileId                     = 0 -- Local ID, not fid on AniDB.

appSettings                    = copyTable(DEFAULT_APP_SETTINGS, true)

updateAvailableMessageReceived = false



dialogs = require"dialogs"



--==============================================================
--= Functions ==================================================
--==============================================================



-- fileInfo = addFileInfo( path )
-- fileInfo = addFileInfo( fileInfo ) -- For the loading phase.
function addFileInfo(pathOrFi)
	local fi, path

	if type(pathOrFi) == "table" then
		fi   = pathOrFi
		path = fi.path

	else
		path = pathOrFi
		fi   = fileInfos[path]
		if fi then  return fi  end

		lastFileId = lastFileId+1

		fi = {
			-- Don't forget to update FILE_INFO_DONT_SERIALIZE when necessary.
			id           = lastFileId,

			path         = path,
			name         = "",
			folder       = "",

			ed2k         = "",
			size         = assert(getFileSize(path)),
			fid          = -1,

			mylistStatus = MYLIST_STATUS_UNKNOWN,
			lid          = -1,

			isHashing    = false,
		}
	end

	fi.name      = getFilename(path)
	fi.folder    = getDirectory(path)
	fi.isHashing = false

	setAndInsert(fileInfos, path, fi)

	if #fileInfos == 1 then
		fileList:DeleteItem(0)
		fileList:Enable(true)
	end

	local wxRow = listCtrlInsertRow(fileList, fi.name, fi.folder, formatBytes(fi.size), getFileViewed(fi), getFileStatus(fi))
	fileList:SetItemData(wxRow, fi.id)

	softUpdateFileInfo(fi)
	return fi
end

function removeFileInfo(fileInfo)
	fileInfo = fileInfos[fileInfo.path]
	if not fileInfo then  return  end

	unsetAndRemove(fileInfos, fileInfo.path)

	local wxRow = getFileRow(fileInfo)
	fileList:DeleteItem(wxRow)

	if noFileInfos() then
		listCtrlInsertRow(fileList, T"message_dropFilesHereInList")
		listCtrlSelectRows(fileList)
		fileList:Enable(false)
	end
end

function removeSelectedFileInfos()
	local fileInfosSelected, wxRows = getSelectedFileInfos()
	if not fileInfosSelected[1] then  return  end

	for _, fileInfo in ipairsr(fileInfosSelected) do
		removeFileInfo(fileInfo)
	end

	saveFileInfos()
	updateFileList()

	if anyFileInfos() then
		listCtrlSelectRows(fileList, {wxRows[1]}, true)
	end
end



function updateFileList()
	if noFileInfos() then return end

	local commonFolder        = ""
	local allInSameFolder     = true
	local filesInCommonFolder = false

	-- Start the common folder value with a path that isn't the shortest.
	if fileInfos[2] then
		for _, fileInfo in ipairs(fileInfos) do
			if #fileInfo.folder > #commonFolder then
				commonFolder = fileInfo.folder
			end
		end
	end

	-- Narrow down the value to the actual common folder.
	for _, fileInfo in ipairs(fileInfos) do
		for ptr = 1, math.max(#fileInfo.folder, #commonFolder) do
			if fileInfo.folder:byte(ptr) ~= commonFolder:byte(ptr) then
				commonFolder        = commonFolder:sub(1, ptr-1)
				allInSameFolder     = false
				filesInCommonFolder = false
				break
			end
		end

		if #commonFolder <= 5 then
			commonFolder = ""
			break
		end

		if fileInfo.folder == commonFolder then
			filesInCommonFolder = true
		end
	end

	-- print(commonFolder, "allSame", allInSameFolder, "inCommon", filesInCommonFolder)

	local ptr    = (findPreviousChar(commonFolder, "/", #commonFolder) or 1)-1
	local prefix = commonFolder:sub(1, ptr)

	if #prefix <= 5 then  prefix = ""  end

	local usePrefix = (appSettings.truncateFolders and prefix ~= "")

	for _, fileInfo in ipairs(fileInfos) do
		listCtrlSetItem(
			fileList,
			getFileRow(fileInfo),
			FILE_COLUMN_FOLDER-1,
			usePrefix and "..."..fileInfo.folder:sub(#prefix+1) or fileInfo.folder
		)
	end

	listCtrlSort(fileList, function(a, b)
		a = itemWith(fileInfos, "id",a)
		b = itemWith(fileInfos, "id",b)
		if a.folder ~= b.folder then
			return a.folder < b.folder and -1 or 1
		end
		if a.name ~= b.name then
			return a.name < b.name and -1 or 1
		end
		return 0 -- Should never happen.
	end)

	local colorStripe1 = wxCOLOUR_WHITE
	local colorStripe2 = wxColour(245, 245, 245)

	for wxRow = 0, fileList.ItemCount-1 do
		local color = (wxRow%2 == 0 and colorStripe1 or colorStripe2)
		fileList:SetItemBackgroundColour(wxRow, color)
	end
end


do
	local FILE_INFO_VERSION = 1

	function saveFileInfos()
		logprint("App", "Saving file info.")

		if DEBUG_DISABLE_VARIOUS_FILE_SAVING then  return  end

		local path = DIR_CACHE.."/files"

		backupFileIfExists(path)
		local file = assert(openFile(path, "w"))

		writeLine(file, FILE_INFO_VERSION)

		for _, fileInfo in ipairs(fileInfos) do
			writeLine(file) -- An empty line separates file entries.

			for k, v in pairsSorted(fileInfo) do
				if not FILE_INFO_DONT_SERIALIZE[k] then
					writeSimpleKv(file, k, v, path)
				end
			end
		end

		file:close()
	end

	function loadFileInfos()
		local path = DIR_CACHE.."/files"
		if not isFile(path) then  return  end

		local file = assert(openFile(path, "r"))
		local ln   = 0

		local fileInfo = nil

		local function finishEntry()
			if not fileInfo then  return  end

			if fileInfo.id then
				addFileInfo(fileInfo)
				lastFileId = math.max(fileInfo.id, lastFileId)
			else
				logprinterror("App", "%s:%d: Missing ID for previous entry. Skipping.", path, ln)
			end

			fileInfo = nil
		end

		lastFileId = 0

		for line in file:lines() do
			ln = ln+1

			if ln == 1 then
				local ver = tonumber(line)
				if not (isInt(ver) and ver >= 1 and ver <= FILE_INFO_VERSION) then
					errorf("%s:%d: %s", path, ln, T"error_badFileVersion")
				end

			elseif line == "" then
				finishEntry() -- An empty line separates file entries.

			else
				local k, v = parseSimpleKv(line, path, ln)

				if k then
					fileInfo = fileInfo or {}
					if fileInfo[k] ~= nil then
						logprinterror("App", "%s:%d: Duplicate key '%s'. Overwriting.", path, ln, k)
					end
					fileInfo[k] = v
				end
			end
		end

		finishEntry()
		file:close()

		updateFileList()
	end
end



-- setFileInfo( fileInfo, key, value [, force=false ] )
function setFileInfo(fileInfo, k, v, force)
	if fileInfo[k] == v and not force then
		return
	end

	if k == "path" then
		local pathOld = fileInfo.path
		local pathNew = v

		fileInfo.path   = pathNew
		fileInfo.name   = getFilename(pathNew)
		fileInfo.folder = getDirectory(pathNew)

		-- @Robustness: Check if fileInfos[pathNew] is already occupied.
		fileInfos[pathOld] = nil
		fileInfos[pathNew] = fileInfo

		anidb:reportLocalFileMoved(pathOld, pathNew)

	else
		fileInfo[k] = v
	end

	local wxRow = getFileRow(fileInfo)
	if not wxRow then
		logprinterror("App", "File %d is not in list.", fileInfo.id)
		return
	end

	if isAny(k, "lid","fid","ed2k","isHashing","mylistStatus") then
		fileList:SetItem(wxRow, FILE_COLUMN_VIEWED-1, getFileViewed(fileInfo))
		fileList:SetItem(wxRow, FILE_COLUMN_STATUS-1, getFileStatus(fileInfo))

	elseif isAny(k, "path") then
		fileList:SetItem(wxRow, FILE_COLUMN_FILE-1,   fileInfo.name)
		fileList:SetItem(wxRow, FILE_COLUMN_FOLDER-1, fileInfo.folder)
	end
end

-- isNowUpdating = softUpdateFileInfo( fileInfo [, refresh=false ] )
function softUpdateFileInfo(fi, refresh)
	if not isFile(fi.path) then
		-- Do nothing here. checkFileInfos() should be called at some point after this function.

	elseif fi.ed2k == "" and appSettings.autoHash then
		if not fi.isHashing then
			setFileInfo(fi, "isHashing", true)
			anidb:hashFile(fi.path)
		end
		return true

	elseif fi.lid == -1 and fi.ed2k ~= "" and not appSettings.autoAddToMylist and fi.mylistStatus == MYLIST_STATUS_UNKNOWN then
		anidb:getMylistByEd2k(fi.ed2k, fi.size)
		return true

	elseif fi.lid == -1 and fi.ed2k ~= "" and appSettings.autoAddToMylist and fi.mylistStatus ~= MYLIST_STATUS_INVALID then
		-- This will act as getMylistByEd2k() if an entry already exist.
		anidb:addMylistByEd2k(fi.ed2k, fi.size)
		return true

	elseif fi.lid ~= -1 and (fi.fid == -1 or refresh) then
		-- The app probably stopped before getting this info last session.
		anidb:getMylist(fi.lid, refresh)
		return true
	end

	return false
end



function getFileStatus(fileInfo)
	return
		nil
		or fileInfo.isHashing                              and T"message_calculatingHash"
		or fileInfo.mylistStatus == MYLIST_STATUS_INVALID  and T"label_notOnAnidb"
		or fileInfo.mylistStatus == MYLIST_STATUS_YES      and T"label_inMylist"
		or fileInfo.mylistStatus == MYLIST_STATUS_NO       and T"label_notInMylist"
		or fileInfo.ed2k         ~= ""                     and T"label_hashed"
		or T"label_notHashed"
end

function getFileViewed(fileInfo)
	local lid = fileInfo.lid
	if lid == -1 then  return "?"  end

	local mylistEntry = anidb:getCacheMylist(lid)
	if not mylistEntry then  return "?"  end

	local time = mylistEntry.viewdate
	return time == 0 and T"label_no" or os.date("%Y-%m-%d", time)
end



function getFileRow(fileInfo)
	local wxRow = fileList:FindItem(-1, fileInfo.id)
	return wxRow ~= -1 and wxRow or nil
end

function getFileInfoByRow(wxRow)
	return (itemWith(fileInfos, "id",fileList:GetItemData(wxRow)))
end



function eachFileInfoByRow(wxRows)
	local i = 0
	return function()
		i = i+1
		local wxRow = wxRows[i]
		if wxRow then  return getFileInfoByRow(wxRow)  end
	end
end

function getFileInfosByRows(wxRows)
	local fileInfosToReturn = {}
	for _, wxRow in ipairs(wxRows) do
		table.insert(fileInfosToReturn, getFileInfoByRow(wxRow))
	end
	return fileInfosToReturn
end

-- fileInfos, wxRows = getSelectedFileInfos( [ onlyHashed=false ] )
function getSelectedFileInfos(onlyHashed)
	local wxRows            = listCtrlGetSelectedRows(fileList)
	local fileInfosSelected = getFileInfosByRows(wxRows)

	if onlyHashed then
		for i, fileInfo in ipairsr(fileInfosSelected) do
			if fileInfo.ed2k == "" then
				table.remove(fileInfosSelected, i)
			end
		end
	end

	return fileInfosSelected, wxRows
end



function anyFileInfos()
	return fileInfos[1] ~= nil
end

function noFileInfos()
	return not fileInfos[1]
end



-- setStatusText( text )
-- setStatusText( format, ... )
function setStatusText(s, ...)
	if select("#", ...) > 0 then
		s = s:format(...)
	end
	topFrame.StatusText = s
end



do
	local PATH_SETTINGS = DIR_CONFIG.."/settings"

	local saveScheduled = false
	local saveTimer     = nil

	local callbacks     = {}

	-- setSetting( key, value    [, startSaveTimer=true ] )
	-- setSetting( key, callback [, startSaveTimer=true ] )
	-- Note: The callback is called twice - immediately and before saving.
	function setSetting(k, vOrCb, startSaveTimer)
		if settingsAreFrozen then  return  end

		local vOld = appSettings[k]
		if vOld == vOrCb then  return  end

		local vOrCbType = type(vOrCb)

		if vOrCbType == "table" and type(vOld) == "table" then
			local t = vOrCb
			if compareTables(t, vOld, true) then  return  end
		end

		if vOrCbType == "function" then
			local cb       = vOrCb
			callbacks[k]   = cb
			appSettings[k] = cb()
		else
			local v        = vOrCb
			callbacks[k]   = nil
			appSettings[k] = v
		end

		if startSaveTimer == false then
			saveScheduled = true
		else
			scheduleSaveSettings()
		end
	end

	function loadSettings()
		readSimpleEntryFile(PATH_SETTINGS, appSettings, true)
	end

	function saveSettings()
		if not saveScheduled then  return  end
		if isPaused()        then  return  end

		local _callbacks = callbacks
		callbacks = {}
		for k, cb in pairs(_callbacks) do
			appSettings[k] = cb()
		end

		if saveTimer then  saveTimer:Stop()  end

		logprint("App", "Saving settings.")

		if not DEBUG_DISABLE_VARIOUS_FILE_SAVING then
			assert(writeSimpleEntryFile(PATH_SETTINGS, appSettings))
		end
		saveScheduled = false
	end

	-- scheduleSaveSettings( [ delay=SAVE_DELAY ] )
	function scheduleSaveSettings(delay)
		delay = delay or SAVE_DELAY

		saveScheduled = true

		saveTimer = saveTimer or newTimer(function(e)  saveSettings()  end)
		saveTimer:Start(delay, true)
	end

	-- scheduleSaveSettingsIfNeeded( )
	function scheduleSaveSettingsIfNeeded()
		if not saveScheduled                   then  return  end
		if saveTimer and saveTimer:IsRunning() then  return  end

		scheduleSaveSettings()
	end

	function setSettingsChanged()
		saveScheduled = true
	end

	function restoreDefaultSettings()
		logprint("App", "Restoring default settings.")

		local langCode = appSettings.language

		appSettings = copyTable(DEFAULT_APP_SETTINGS, true)
		appSettings.language = langCode

		anidb:resetSettings()

		scheduleSaveSettings()
	end
end




do
	local function silentlyCheckFileAndMaybeRegisterMove(fileInfo, dirNew)
		local pathOld = fileInfo.path
		if isFile(pathOld) then  return  end

		local pathNew = F("%s/%s", dirNew, fileInfo.name)

		if not fileInfos[pathNew] and isFile(pathNew) and getFileSize(pathNew) == fileInfo.size then
			setFileInfo(fileInfo, "path", pathNew)
		end
	end

	local function checkFile(fileInfo, i)
		local pathOld = fileInfo.path
		if isFile(pathOld) then  return false  end

		if appSettings.autoRemoveDeletedFiles then
			-- @Incomplete: Add a setting to also mark removed files as 'deleted' in MyList,
			-- like the user can when autoRemoveDeletedFiles is off. (Not sure if this is
			-- such a good idea. Seems error-prone.)
			removeFileInfo(fileInfo)
			anidb:reportLocalFileDeleted(pathOld)
			return true
		end

		local showMylistOptions           = (fileInfo.lid ~= -1)
		local pathNew, setMylistToDeleted = dialogs.missingFile(pathOld, showMylistOptions)

		if not pathNew then
			-- Leave fileInfo as-is. (Should we mark it so we don't ask for a new path again? @UX)
			return false

		elseif pathNew == "" then
			if showMylistOptions and setMylistToDeleted then
				anidb:editMylist(fileInfo.lid, {state=MYLIST_STATE_DELETED})
				-- @Robustness: Don't do the below stuff until the update on AniDB is confirmed.
			end

			removeFileInfo(fileInfo)
			anidb:reportLocalFileDeleted(pathOld)

			return true
		end

		local fileSizeNew = getFileSize(pathNew)

		if fileSizeNew ~= fileInfo.size then
			showError(
				T"label_fileIsDifferent",
				F(
					"%s\n\n%s\n%s\n\n%s",
					T("error_fileSizeMismatch1"),
					T("error_fileSizeMismatch2", {n=F("%.0f", fileInfo.size)}),
					T("error_fileSizeMismatch3", {n=F("%.0f", fileSizeNew)}),
					pathNew
				)
			)

			return checkFile(fileInfo, i)
		end

		setFileInfo(fileInfo, "path", pathNew)

		for j = 1, i-1 do
			silentlyCheckFileAndMaybeRegisterMove(fileInfos[j], getDirectory(pathNew))
		end

		return true
	end

	function checkFileInfos()
		local anyFileInfoChanged = false

		-- Check deleted/moved files.
		for i, fileInfo in ipairsr(fileInfos) do
			anyFileInfoChanged = checkFile(fileInfo, i) or anyFileInfoChanged
		end

		-- @Incomplete: Do autoHash etc.

		if anyFileInfoChanged then
			saveFileInfos()
			updateFileList()
		end
	end
end



do
	local pauseKeys = {}

	function pause(k)
		if pauseKeys[k] then
			pauseKeys[k] = pauseKeys[k]+1

		else
			if not isPaused() then
				logprint(nil, "Paused by '%s'.", k)
			end
			pauseKeys[k] = 1
		end
	end

	function unpause(k)
		if not pauseKeys[k] then
			logprinterror(nil, "Tried to remove non-existent pause key '%s'.", k)

		elseif pauseKeys[k] > 1 then
			pauseKeys[k] = pauseKeys[k]-1

		else
			pauseKeys[k] = nil
			if not isPaused() then
				logprint(nil, "Unpaused by '%s'.", k)
				scheduleSaveSettingsIfNeeded()
			end
		end
	end

	function isPaused()
		return next(pauseKeys) ~= nil
	end
end



function openContainingFolder(path)
	if isDirectory(getDirectory(path)) then
		showFileInExplorer(path)
	else
		showError("Error", F("%s\n\n%s", T"error_missingDirectory", path))
		checkFileInfos()
	end
end


